/**********************************************************************
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Contributors:
 *    IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.wtp.releng.tools.component.api;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.eclipse.jdt.core.util.ClassFormatException;
import org.eclipse.jdt.core.util.IClassFileReader;
import org.eclipse.jdt.core.util.IExceptionAttribute;
import org.eclipse.jdt.core.util.IFieldInfo;
import org.eclipse.jdt.core.util.IMethodInfo;
import org.eclipse.jdt.internal.core.util.ClassFileReader;
import org.eclipse.wtp.releng.tools.component.ILocation;
import org.eclipse.wtp.releng.tools.component.ILocationVisitor;
import org.eclipse.wtp.releng.tools.component.classes.IClassVisitor;
import org.eclipse.wtp.releng.tools.component.classes.LibVisitor;
import org.eclipse.wtp.releng.tools.component.images.ImagesUtil;
import org.eclipse.wtp.releng.tools.component.internal.FileLocation;
import org.eclipse.wtp.releng.tools.component.internal.Location;
import org.eclipse.wtp.releng.tools.component.model.ComponentXML;
import org.eclipse.wtp.releng.tools.component.model.Package;
import org.eclipse.wtp.releng.tools.component.model.Plugin;
import org.eclipse.wtp.releng.tools.component.model.Type;
import org.eclipse.wtp.releng.tools.component.util.CommandOptionParser;
import org.eclipse.wtp.releng.tools.component.xsl.XSLUtil;

public class API2ComponentAPI implements IClassVisitor
{
  private String api;
  private String src;
  private String outputDir;
  private Collection includes;
  private Collection excludes;
  private boolean html;
  private boolean readInterface = false;
  private boolean skipAPIGen = false;
  private Map interface2ImplClasses = new HashMap(0);

  public String getApi()
  {
    return api;
  }

  public void setApi(String api)
  {
    this.api = api;
  }

  public String getSrc()
  {
    return src;
  }

  public void setSrc(String src)
  {
    this.src = src;
  }

  public String getOutputDir()
  {
    return outputDir;
  }

  public void setOutputDir(String outputDir)
  {
    this.outputDir = addTrailingSeperator(outputDir);
  }

  public Collection getIncludes()
  {
    return includes;
  }

  public void setIncludes(Collection includes)
  {
    this.includes = includes;
  }

  public Collection getExcludes()
  {
    return excludes;
  }

  public void setExcludes(Collection excludes)
  {
    this.excludes = excludes;
  }

  public boolean isHtml()
  {
    return html;
  }

  public void setHtml(boolean html)
  {
    this.html = html;
  }

  public boolean isReadInterface()
  {
    return readInterface;
  }

  public void setReadInterface(boolean readInterface)
  {
    this.readInterface = readInterface;
  }

  public boolean isSkipAPIGen()
  {
    return skipAPIGen;
  }

  public void setSkipAPIGen(boolean skipAPIGen)
  {
    this.skipAPIGen = skipAPIGen;
  }

  public List getImplClasses(String interfaceName)
  {
    List implClasses = (List)interface2ImplClasses.get(interfaceName);
    return implClasses != null ? new ArrayList(implClasses) : new ArrayList(0);
  }

  private Map pluginId2CompXML = new HashMap();

  public void execute()
  {
    // Collection component.xml files
    ILocation apiLocation = Location.createLocation(new File(api));
    ComponentXMLVisitor compXMLVisitor = new ComponentXMLVisitor();
    apiLocation.accept(compXMLVisitor);
    for (Iterator it = compXMLVisitor.getCompXMLs().iterator(); it.hasNext();)
    {
      ComponentXML compXML = (ComponentXML)it.next();
      for (Iterator it2 = compXML.getPlugins().iterator(); it2.hasNext();)
      {
        pluginId2CompXML.put(((Plugin)it2.next()).getId(), compXML);
      }
    }
    // Generate api-info.xml files
    ILocation srcLocation = Location.createLocation(new File(src));
    LibVisitor libVisitor = new LibVisitor();
    srcLocation.accept(libVisitor);
    libVisitor.setClassVisitor(this);
    srcLocation.accept(libVisitor);
    try
    {
      if (cachedCompAPI != null)
        cachedCompAPI.save();
    }
    catch (IOException e)
    {
      throw new RuntimeException(e);
    }
    // Generate HTML
    if (isHtml())
      genHTML();
  }

  private void genHTML()
  {
    ImagesUtil.copyAll(outputDir);
    final StringBuffer summary = new StringBuffer();
    summary.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    summary.append("<root>");
    ILocation outputLoc = Location.createLocation(new File(outputDir));
    outputLoc.accept(new ILocationVisitor()
    {
      public boolean accept(ILocation location)
      {
        if (location.getName().endsWith("api-info.xml"))
        {
          summary.append("<api-info file=\"");
          summary.append(location.getAbsolutePath().substring(outputDir.length()));
          summary.append("\"/>");
        }
        return true;
      }
    });
    summary.append("</root>");
    try
    {
      XSLUtil.transform
      (
        ClassLoader.getSystemResourceAsStream("org/eclipse/wtp/releng/tools/component/xsl/api-info-summary.xsl"),
        new ByteArrayInputStream(summary.toString().getBytes()),
        new FileOutputStream(new File(outputDir + "/api-info-summary.html")),
        outputDir
      );
    }
    catch (Throwable e)
    {
      e.printStackTrace();
    }
  }

  public boolean visit(String pluginId, ILocation classLoc)
  {
    String className = classLoc.getName();
    className = className.substring(0, className.length() - ".class".length());
    className = className.replace('/', '.');
    className = className.replace('\\', '.');
    if (include(className))
    {
      int i = className.lastIndexOf('.');
      String packageName = (i != -1) ? className.substring(0, i) : "";
      String localName = (i != -1) ? className.substring(i + 1) : className;
      boolean isAPI;
      ComponentXML compXML = (ComponentXML)pluginId2CompXML.get(pluginId);
      if (compXML != null)
      {
        Package pkg = compXML.getPackage(packageName);
        if (pkg != null)
        {
          Type type = pkg.getType(localName);
          if (type != null)
          {
            if (!type.isReference() && !type.isSubclass() && !type.isImplement() && !type.isInstantiate())
            {
              isAPI = false;
            }
            else
            {
              isAPI = true;
            }
          }
          else
          {
            isAPI = pkg.isApi();
          }
        }
        else
        {
          isAPI = false;
        }
      }
      else
      {
        isAPI = false;
      }
      IClassFileReader reader = null;
      if (readInterface)
      {
        try
        {
          reader = read(classLoc);
          if (!reader.isInterface())
          {
            char[][] names = reader.getInterfaceNames();
            for (int j = 0; j < names.length; j++)
            {
              String interfaceName = new String(names[j]).replace('/', '.');
              List implClasses = (List)interface2ImplClasses.get(interfaceName);
              if (implClasses == null)
              {
                implClasses = new ArrayList(1);
                interface2ImplClasses.put(interfaceName, implClasses);
              }
              implClasses.add(className);
            }
          }
        }
        catch (IOException ioe)
        {
          ioe.printStackTrace();
        }
        catch (ClassFormatException cfe)
        {
          cfe.printStackTrace();
        }
      }
      if (!isSkipAPIGen() && isAPI)
      {
        try
        {
          ComponentAPI compAPI = getComponentAPI(pluginId);
          PackageAPI pkgAPI = (PackageAPI)compAPI.getPackageAPI(packageName);
          if (pkgAPI == null)
          {
            pkgAPI = new PackageAPI();
            pkgAPI.setName(packageName);
            compAPI.addPackageAPI(pkgAPI);
          }
          ClassAPI classAPI = pkgAPI.getClassAPI(localName);
          if (classAPI == null)
          {
            if (reader == null)
              reader = read(classLoc);
            classAPI = new ClassAPI();
            classAPI.setName(localName);
            classAPI.setAccess(reader.getAccessFlags());
            classAPI.setSuperClass(new String(reader.getSuperclassName()).replace('/', '.'));
            pkgAPI.addClassAPI(classAPI);
            IMethodInfo[] methods = reader.getMethodInfos();
            for (int j = 0; j < methods.length; j++)
            {
              MethodAPI methodAPI = new MethodAPI();
              methodAPI.setName(new String(methods[j].getName()));
              methodAPI.setAccess(methods[j].getAccessFlags());
              methodAPI.setDescriptor(new String(methods[j].getDescriptor()));
              IExceptionAttribute exceptionAttr = methods[j].getExceptionAttribute();
              if (exceptionAttr != null)
              {
                char[][] exceptionNames = exceptionAttr.getExceptionNames();
                List exceptions = new ArrayList(exceptionNames.length);
                for (int k = 0; k < exceptionNames.length; k++)
                {
                  exceptions.add(new String(exceptionNames[k]).replace('/', '.'));
                }
                methodAPI.addThrows(exceptions);
              }
              classAPI.addMethodAPI(methodAPI);
            }
            IFieldInfo[] fields = reader.getFieldInfos();
            for (int j = 0; j < fields.length; j++)
            {
              FieldAPI fieldAPI = new FieldAPI();
              fieldAPI.setName(new String(fields[j].getName()));
              fieldAPI.setAccess(fields[j].getAccessFlags());
              fieldAPI.setDescriptor(new String(fields[j].getDescriptor()));
              classAPI.addFieldAPI(fieldAPI);
            }
          }
        }
        catch (IOException e)
        {
          throw new RuntimeException(e);
        }
        catch (ClassFormatException e)
        {
          throw new RuntimeException(e);
        }
      }
    }
    return true;
  }

  private IClassFileReader read(ILocation classLoc) throws IOException, ClassFormatException
  {
    InputStream is = null;
    ByteArrayOutputStream baos = null;
    try
    {
      byte[] b = new byte[8192];
      baos = new ByteArrayOutputStream(8192);
      is = classLoc.getInputStream();
      for (int read = is.read(b); read != -1; read = is.read(b))
      {
        baos.write(b, 0, read);
      }
      is.close();
      baos.close();
      return new ClassFileReader(baos.toByteArray(), IClassFileReader.CONSTANT_POOL | IClassFileReader.METHOD_INFOS | IClassFileReader.FIELD_INFOS | IClassFileReader.SUPER_INTERFACES);
    }
    finally
    {
      if (is != null)
      {
        try
        {
          is.close();
        }
        catch (IOException e)
        {
        }
      }
      if (baos != null)
      {
        try
        {
          baos.close();
        }
        catch (IOException e)
        {
        }
      }
    }
  }

  private boolean include(String name)
  {
    name = name.replace('/', '.');
    name = name.replace('\\', '.');
    if (excludes != null && !excludes.isEmpty())
      for (Iterator it = excludes.iterator(); it.hasNext();)
        if (name.matches((String)it.next()))
          return false;
    if (includes != null && !includes.isEmpty())
    {
      for (Iterator it = includes.iterator(); it.hasNext();)
        if (name.matches((String)it.next()))
          return true;
      return false;
    }
    return true;
  }
  private ComponentAPI cachedCompAPI;

  private ComponentAPI getComponentAPI(String id) throws IOException
  {
    if (cachedCompAPI != null)
    {
      if (cachedCompAPI.getName().equals(id))
      {
        return cachedCompAPI;
      }
      else
      {
        cachedCompAPI.save();
      }
    }
    StringBuffer sb = new StringBuffer(outputDir);
    sb.append(id);
    sb.append("/api-info.xml");
    File file = new File(sb.toString());
    cachedCompAPI = new ComponentAPI();
    cachedCompAPI.setName(id);
    cachedCompAPI.setLocation(new FileLocation(file));
    if (file.exists())
      cachedCompAPI.load();
    return cachedCompAPI;
  }

  protected String addTrailingSeperator(String s)
  {
    if (s != null && !s.endsWith("/") && !s.endsWith("\\"))
    {
      StringBuffer sb = new StringBuffer(s);
      sb.append('/');
      return sb.toString();
    }
    else
    {
      return s;
    }
  }

  public static void main(String[] args)
  {
    CommandOptionParser optionParser = new CommandOptionParser(args);
    Map options = optionParser.getOptions();
    Collection api = (Collection)options.get("api");
    Collection src = (Collection)options.get("src");
    Collection outputDir = (Collection)options.get("outputDir");
    Collection includes = (Collection)options.get("includes");
    Collection excludes = (Collection)options.get("excludes");
    Collection html = (Collection)options.get("html");
    if (api == null || src == null || outputDir == null || api.isEmpty() || src.isEmpty() || outputDir.isEmpty())
    {
      printUsage();
      System.exit(-1);
    }
    API2ComponentAPI api2CompAPI = new API2ComponentAPI();
    api2CompAPI.setApi((String)api.iterator().next());
    api2CompAPI.setSrc((String)src.iterator().next());
    api2CompAPI.setOutputDir((String)outputDir.iterator().next());
    api2CompAPI.setIncludes(includes);
    api2CompAPI.setExcludes(excludes);
    api2CompAPI.setHtml(html != null);
    api2CompAPI.execute();
  }

  private static void printUsage()
  {
    System.out.println("Usage: java org.eclipse.wtp.releng.tools.component.api.API2ComponentAPI -api <api> -src <src> -outputDir <outputDir> [-options]");
    System.out.println("");
    System.out.println("\t-api\t\t<api>\t\tlocation of your component.xml");
    System.out.println("\t-src\t\t<src>\t\tlocation of a Eclipse-based product");
    System.out.println("\t-outputDir\t<outputDir>\toutput directory of component.xml files");
    System.out.println("");
    System.out.println("where options include:");
    System.out.println("");
    System.out.println("\t-includes\t<includes>\tspace seperated packages to include");
    System.out.println("\t-excludes\t<excludes>\tspace seperated packages to exclude");
    System.out.println("\t-html\t\t\t\t\tgenerate HTML output");
  }
}