/**********************************************************************
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Contributors:
 *    IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.wtp.releng.tools.component.java;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.Javadoc;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodRef;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.PackageDeclaration;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.TagElement;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.wtp.releng.tools.component.api.ClassAPI;
import org.eclipse.wtp.releng.tools.component.api.ComponentAPI;
import org.eclipse.wtp.releng.tools.component.api.FieldAPI;
import org.eclipse.wtp.releng.tools.component.api.JavadocCoverage;
import org.eclipse.wtp.releng.tools.component.api.MethodAPI;
import org.eclipse.wtp.releng.tools.component.api.PackageAPI;

public class JavadocVisitor extends ASTVisitor
{
  private ComponentAPI compAPI;
  private String packageName;
  private Map imports;
  private ClassAPI classAPI;

  public JavadocVisitor(ComponentAPI compAPI)
  {
    super(true);
    this.compAPI = compAPI;
    this.packageName = null;
    this.imports = new HashMap();
    this.classAPI = null;
  }

  public boolean visit(PackageDeclaration node)
  {
    packageName = node.getName().getFullyQualifiedName();
    return true;
  }

  public boolean visit(ImportDeclaration node)
  {
    String fullName = node.getName().getFullyQualifiedName();
    int i = fullName.lastIndexOf('.');
    String localName = (i != -1) ? fullName.substring(i + 1) : fullName;
    imports.put(localName, fullName);
    return true;
  }

  public boolean visit(TypeDeclaration node)
  {
    boolean isPrivate = Modifier.isPrivate(node.getModifiers());
    if (!isPrivate && isAPI(node))
    {
      PackageAPI pkgAPI = compAPI.getPackageAPI(packageName);
      if (pkgAPI == null)
      {
        pkgAPI = new PackageAPI();
        pkgAPI.setName(packageName);
        compAPI.addPackageAPI(pkgAPI);
      }
      String className = node.getName().getFullyQualifiedName();
      classAPI = pkgAPI.getClassAPI(className);
      if (classAPI == null)
      {
        classAPI = new ClassAPI();
        classAPI.setName(className);
        classAPI.setAccess(node.getModifiers());
        pkgAPI.addClassAPI(classAPI);
      }
      Javadoc javadoc = node.getJavadoc();
      if (javadoc != null)
      {
        for (Iterator it = javadoc.tags().iterator(); it.hasNext();)
        {
          TagElement tag = (TagElement)it.next();
          String tagName = tag.getTagName();
          if (tagName != null && tagName.equals(TagElement.TAG_SINCE))
          {
            // do nothing because default is true
            return true;
          }
        }
      }
      classAPI.getJavadocCoverage().setHasSince(Boolean.FALSE);
      return true;
    }
    return false;
  }

  private boolean isAPI(TypeDeclaration node)
  {
    return true;
  }

  public boolean visit(MethodDeclaration node)
  {
    if (classAPI != null && !Modifier.isPrivate(node.getModifiers()))
    {
      String methodName = node.isConstructor() ? "&lt;init>" : node.getName().getFullyQualifiedName();
      StringBuffer descriptor = new StringBuffer();

      // input params
      descriptor.append(Signature.C_PARAM_START);
      List params = new ArrayList();
      for (Iterator it = node.parameters().iterator(); it.hasNext();)
      {
        SingleVariableDeclaration param = (SingleVariableDeclaration)it.next();
        params.add(param.getName().getIdentifier());
        descriptor.append(getDescriptor(param.getType()));
      }
      descriptor.append(Signature.C_PARAM_END);

      // return type
      boolean hasReturn = false;
      org.eclipse.jdt.core.dom.Type returnType = node.getReturnType2();
      if (returnType instanceof PrimitiveType)
        if (((PrimitiveType)returnType).getPrimitiveTypeCode() == PrimitiveType.VOID)
          hasReturn = true;
      descriptor.append(getDescriptor(returnType));

      // exceptions
      List thrownExceptions = new ArrayList();
      for (Iterator it = node.thrownExceptions().iterator(); it.hasNext();)
      {
        Name name = (Name)it.next();
        thrownExceptions.add(name.getFullyQualifiedName());
      }

      // javadoc
      boolean hasJavadoc = false;
      Javadoc javadoc = node.getJavadoc();
      if (javadoc != null)
      {
        hasJavadoc = true;
        for (Iterator it = javadoc.tags().iterator(); it.hasNext();)
        {
          TagElement tag = (TagElement)it.next();
          String tagName = tag.getTagName();
          if (tagName != null)
          {
            if (tagName.equals(TagElement.TAG_RETURN))
              hasReturn = tag.fragments().size() > 0;
            else if (tagName.equals(TagElement.TAG_PARAM))
            {
              List fragments = tag.fragments();
              if (fragments.size() > 1)
              {
                Object fragment = fragments.get(0);
                if (fragment instanceof SimpleName)
                {
                  params.remove(((SimpleName)fragment).getIdentifier());
                }
              }
            }
            else if (tagName.equals(TagElement.TAG_THROWS) || tagName.equals(TagElement.TAG_EXCEPTION))
            {
              List fragments = tag.fragments();
              if (fragments.size() > 1)
              {
                Object fragment = fragments.get(0);
                if (fragment instanceof Name)
                {
                  thrownExceptions.remove(((Name)fragment).getFullyQualifiedName());
                }
              }
            }
            else if (tagName.equals(TagElement.TAG_SEE))
            {
              List fragments = tag.fragments();
              if (fragments.size() > 0)
              {
                Object fragment = fragments.get(0);
                if (fragment instanceof MethodRef)
                {
                  MethodRef methodRef = (MethodRef)fragment;
                  if (methodName.equals(methodRef.getName().getFullyQualifiedName()))
                  {
                    return true;
                  }
                }
              }
            }
          }
        }
      }
      String methodDescriptor = descriptor.toString();
      MethodAPI methodAPI = classAPI.getMethodAPI(methodName, methodDescriptor);
      if (methodAPI == null)
      {
        methodAPI = new MethodAPI();
        methodAPI.setName(methodName);
        methodAPI.setAccess(node.getModifiers());
        methodAPI.setDescriptor(methodDescriptor);
        classAPI.addMethodAPI(methodAPI);
      }
      if (!hasJavadoc || !hasReturn || params.size() > 0 || thrownExceptions.size() > 0)
      {
        JavadocCoverage javadocCoverage = methodAPI.getJavadocCoverage();
        if (!hasJavadoc)
        {
          javadocCoverage.setHasDoc(Boolean.FALSE);
        }
        if (!hasReturn)
        {
          javadocCoverage.setHasReturn(Boolean.FALSE);
        }
        if (params.size() > 0)
        {
          for (Iterator it = params.iterator(); it.hasNext();)
          {
            javadocCoverage.addMissingParam((String)it.next());
          }
        }
        if (thrownExceptions.size() > 0)
        {
          for (Iterator it = thrownExceptions.iterator(); it.hasNext();)
          {
            javadocCoverage.addMissingThrow((String)it.next());
          }
        }
      }
    }
    return true;
  }

  private String getDescriptor(org.eclipse.jdt.core.dom.Type type)
  {
    if (type != null)
    {
      String descriptor;
      if (type.isPrimitiveType())
      {
        descriptor = Signature.createTypeSignature(type.toString(), true);
      }
      else if (type.isArrayType())
      {
        String s = type.toString();
        int i = s.lastIndexOf('[');
        StringBuffer sb = new StringBuffer();
        if (i != -1)
        {
          String s2 = s.substring(0, i);
          if (isPrimitiveType(s2))
          {
            sb.append(s);
          }
          else
          {
            sb.append(resolveFromImports(s2));
            sb.append(s.substring(i));
          }
        }
        else
        {
          sb.append(s);
        }
        descriptor = Signature.createTypeSignature(sb.toString(), true);
      }
      else
      {
        descriptor = Signature.createTypeSignature(resolveFromImports(type.toString()), true);
      }
      return descriptor.replace('.', '/');
    }
    else
    {
      StringBuffer sb = new StringBuffer();
      sb.append(Signature.C_VOID);
      return sb.toString();
    }
  }

  private boolean isPrimitiveType(String type)
  {
    return type.equals("byte") ||
      type.equals("short") ||
      type.equals("char") ||
      type.equals("int") ||
      type.equals("long") ||
      type.equals("float") ||
      type.equals("double") ||
      type.equals("boolean");
  }

  private String resolveFromImports(String name)
  {
    String resolvedName = (String)imports.get(name);
    if (resolvedName != null)
    {
      return resolvedName;
    }
    else
    {
      StringBuffer sb = new StringBuffer();
      sb.append("java.lang.");
      sb.append(name);
      try
      {
        Class.forName(sb.toString());
        imports.put(name, sb.toString());
        return sb.toString();
      }
      catch (ClassNotFoundException e)
      {
      }
      sb = new StringBuffer();
      sb.append(packageName);
      sb.append('.');
      sb.append(name);
      imports.put(name, sb.toString());
      return sb.toString();
    }
  }

  public boolean visit(FieldDeclaration node)
  {
    if (classAPI != null && !Modifier.isPrivate(node.getModifiers()))
    {
      VariableDeclarationFragment varDeclFragment = (VariableDeclarationFragment)node.fragments().iterator().next();
      String fieldName = varDeclFragment.getName().getFullyQualifiedName();
      FieldAPI fieldAPI = classAPI.getFieldAPI(fieldName);
      if (fieldAPI == null)
      {
        fieldAPI = new FieldAPI();
        fieldAPI.setName(fieldName);
        fieldAPI.setAccess(node.getModifiers());
        fieldAPI.setDescriptor(getDescriptor(node.getType()));
        classAPI.addFieldAPI(fieldAPI);
      }
      if (node.getJavadoc() == null)
      {
        fieldAPI.setJavadocCoverage(new JavadocCoverage());
      }
    }
    return true;
  }
}